# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

PACKAGE_PREFIX = '_packages'
MATERIALIZED_DATA_PREFIX='_materialized_data'